/* 
  Copyright (c) 2011 Stephen Chamberlin.  All rights reserved.
  Backcountry Logger by Stephen Chamberlin is licensed under a 
  Creative Commons Attribution-NonCommercial-ShareAlike 3.0 Unported License. (CC BY-NC-SA 3.0)
  The terms of the license may be viewed at http://creativecommons.org/licenses/by-nc-sa/3.0/
  Based on a work at www.bigmessowires.com.
  Permissions beyond the scope of this license may be available at www.bigmessowires.com.
*/ 

#ifdef SHAKE_SENSOR

#include <avr/io.h>
#include <avr/interrupt.h>
#include "shake.h"
#include "speaker.h"

// port C
#define SHAKE_PORT PORTC
#define SHAKE_DDR DDRC
#define SHAKE_PIN_IN PC0
#define SHAKE_PIN_OUT PC1

#define SHAKE_TRIP_START_TIMEOUT 5
#define SHAKE_TRIP_STOP_TIMEOUT 5

static volatile uint8_t ignoreOneShake = 0;
static volatile uint8_t isMoving = 0;
static volatile uint8_t shakeDetected = 0;
static volatile uint8_t consecutiveMinutesShaking = 0;
static volatile uint8_t consecutiveMinutesStopped = 0;
static volatile uint16_t tripMinutes = 0;

void ShakeInit()
{
	// enable the internal pull-up resistors for shake-in	
	SHAKE_PORT |= (1<<SHAKE_PIN_IN); 
	
	// set shake-out as output
	SHAKE_DDR |= (1<<SHAKE_PIN_OUT);
		
	// set the pin change interrupt mask		
	PCMSK1 |= (1<<PCINT8);
	
	ShakeEnable(1);
}

void ShakeEnable(uint8_t enable)
{
	if (enable)
	{
		ignoreOneShake = 1;
		
		// drive 0
		SHAKE_PORT &= ~(1<<SHAKE_PIN_OUT);
		// clear interrupt flag
		PCIFR &= ~(1<<PCIF1);
		// enable the pin change interrupts for shake sensor	
		PCICR |= (1<<PCIE1);
	}
	else
	{
		// disable the pin change interrupts for shake sensor		
		PCICR &= ~(1<<PCIE1);
		// drive 1 - uses no power regardless of sensor state
		SHAKE_PORT |= (1<<SHAKE_PIN_OUT);			
	}
}

uint16_t ShakeGetTripTime()
{
	return tripMinutes;
}

uint8_t ShakeIsMoving()
{
	return isMoving;
}

void ShakeUpdate()
{
	// called once per minute
	if (shakeDetected)
	{
		consecutiveMinutesShaking++;
		consecutiveMinutesStopped = 0;
	}
	else
	{
		consecutiveMinutesShaking = 0;
		consecutiveMinutesStopped++;
	}
	
	if (!isMoving)
	{
		if (consecutiveMinutesShaking == SHAKE_TRIP_START_TIMEOUT)
		{
			isMoving = 1;
			tripMinutes = consecutiveMinutesShaking;
			SpeakerBeep(BEEP_TIMER_START);
		}
	}
	else
	{
		// isMoving
		if (consecutiveMinutesStopped == SHAKE_TRIP_STOP_TIMEOUT)
		{
			isMoving = 0;
			SpeakerBeep(BEEP_TIMER_END);
		}
		else
		{
			tripMinutes++;
		}
	}
	
	ShakeEnable(1); // re-enable sensor at the start of each minute
	shakeDetected = 0;
}

// shake sensor state change interrupt
ISR(PCINT1_vect) 
{ 
	if (ignoreOneShake)
	{
		ignoreOneShake = 0;
		return;
	}
	
	shakeDetected = 1;
	
	// disable sensor until explicitly re-enabled, to save power
	ShakeEnable(0);
}	

#endif
